from ..abstract_transformation import *
from ..tasks import *
import numpy as np
import re
from pattern.en import wordnet
import nltk
from nltk.corpus import stopwords
import spacy
import en_core_web_sm

try:
    nltk.data.find('corpora/stopwords')
except LookupError:
    nltk.download('stopwords')

class ChangeSynse(AbstractTransformation):
    """
    Replaces a specified number of random words a string
    with synses from wordnet. Also supports part-of-speech (pos)
    tagging via spaCy to get more natural replacements. 
    """
    def __init__(self, synse='synonym', num_to_replace=-1, return_metadata=False):
        """
        Initializes the transformation and provides an
        opporunity to supply a configuration if needed

        Parameters
        ----------
        synse : str
            The type of word replacement to make.
            Currently support 'synonym', 'antonym', 'hyponym', 
            and 'hypernym' 
        num_to_replace : int
            The number of words randomly selected from the input 
            to replace with synonyms (excludes stop words).
            If -1, then replace as many as possible.
        return_metadata : bool
            whether or not to return metadata, e.g. 
            whether a transform was successfully
            applied or not
        """
        super().__init__() 
        self.synse = synse
        self.synses = {
            'synonym' : all_possible_synonyms,
            'antonym' : all_possible_antonyms,
            'hyponym' : all_possible_hyponyms,
            'hypernym' : all_possible_hypernyms,
        }
        if synse not in self.synses:
            raise ValueError("Invalid synse argument. \n \
                Please select one of the following: 'synonym', 'antonym', 'hyponym', 'hypernym'")
        self.synse_fn = self.synses[self.synse]
        self.num_to_replace = np.Inf if num_to_replace == -1 else num_to_replace
        self.nlp = en_core_web_sm.load()
        # stopwords 
        useful_stopwords = ['few', 'more', 'most', 'against']
        self.stopwords = [w for w in stopwords.words('english') if w not in useful_stopwords]
        self.return_metadata = return_metadata
        self.task_configs = [
            SentimentAnalysis(),
            TopicClassification(),
            Grammaticality(tran_type='SIB'),
            Similarity(input_idx=[1,0], tran_type='INV'),
            Similarity(input_idx=[0,1], tran_type='INV'),
            Similarity(input_idx=[1,1], tran_type='INV'),
            Entailment(input_idx=[1,0], tran_type='INV'),
            Entailment(input_idx=[0,1], tran_type='INV'),
            Entailment(input_idx=[1,1], tran_type='INV'),
        ]
    
    def __call__(self, in_text):
        """Replaces words with synses

        Parameters
        ----------
        in_text : str
            Input string

        Returns
        -------
        out_text : str
            Output string with synses replaced.
        """
        doc = self.nlp(in_text)
        new_words = in_text.split(' ').copy()
        random_word_list = list(set(
            [token.text for token in doc if strip_punct(token.text.lower()) not in self.stopwords]))
        self.np_random.shuffle(random_word_list)
        num_replaced = 0
        for random_word in random_word_list:
            pos = [token.pos_ for token in doc if random_word in token.text][0]
            if pos in ['PROPN']:
                continue
            if pos not in ['NOUN', 'VERB', 'ADJ', 'ADV']:
                pos = None
            options = self.synse_fn(strip_punct(random_word), pos)
            if len(options) >= 1:
                option = self.np_random.choice(list(options))
                new_words = [option if word == random_word else word for word in new_words]
                num_replaced += 1
            if num_replaced >= self.num_to_replace:
                break
        out_text = ' '.join(new_words)
        return out_text

    def get_tran_types(self, task_name=None, tran_type=None, label_type=None):
        pass

    def transform_Xy(self, X, y):

        # transform X
        if isinstance(X, str):
            X = [X]

        assert len(X) == len(self.task_config['input_idx']), ("The number of inputs does not match the expected "
                                                         "amount of {} for the {} task".format(
                                                            self.task_config['input_idx'],
                                                            self.task_config['task_name']))

        X_out = []
        for i, x in zip(self.task_config['input_idx'], X):
            if i == 0:
                X_out.append(x)
                continue
            X_out.append(self(x))

        metadata = {'change': X != X_out}
        X_out = X_out[0] if len(X_out) == 1 else X_out

        # transform y
        if self.task_config['tran_type'] == 'INV':
            y_out = y
        else:
            soften = self.task_config['label_type'] == 'soft'
            if self.task_config['task_name'] == 'grammaticality':
                # hard code for now... :(
                # 0 = ungrammatical, 1 = grammatical
                if isinstance(y, int):
                    if y == 0:
                        y_out = y
                    else: 
                        y_out = invert_label(y, soften=soften)
                else:
                    if np.argmax(y) == 0:
                        y_out = y
                    else: 
                        y_out = invert_label(y, soften=soften)
            else:
               y_out = invert_label(y, soften=soften)
        
        if self.return_metadata: 
            return X_out, y_out, metadata
        return X_out, y_out

class ChangeSynonym(ChangeSynse):
    def __init__(self, num_to_replace=-1, return_metadata=False):
        super().__init__(synse='synonym', num_to_replace=num_to_replace, return_metadata=False)
        self.return_metadata = return_metadata
        self.task_configs = [
            SentimentAnalysis(),
            TopicClassification(),
            Grammaticality(tran_type='SIB'),
            Similarity(input_idx=[1,0], tran_type='INV'),
            Similarity(input_idx=[0,1], tran_type='INV'),
            Similarity(input_idx=[1,1], tran_type='INV'),
            Entailment(input_idx=[1,0], tran_type='INV'),
            Entailment(input_idx=[0,1], tran_type='INV'),
            Entailment(input_idx=[1,1], tran_type='INV'),
        ]

    def __call__(self, in_text):
        return super().__call__(in_text)

    def get_task_configs(self, task_name=None, tran_type=None, label_type=None):
        init_configs = [task() for task in self.task_configs]
        df = self._get_task_configs(init_configs, task_name, tran_type, label_type)
        return df

    def transform_Xy(self, X, y):

        # transform X
        if isinstance(X, str):
            X = [X]

        assert len(X) == len(self.task_config['input_idx']), ("The number of inputs does not match the expected "
                                                         "amount of {} for the {} task".format(
                                                            self.task_config['input_idx'],
                                                            self.task_config['task_name']))

        X_out = []
        for i, x in zip(self.task_config['input_idx'], X):
            if i == 0:
                X_out.append(x)
                continue
            X_out.append(self(x))

        metadata = {'change': X != X_out}
        X_out = X_out[0] if len(X_out) == 1 else X_out

        # transform y
        if self.task_config['tran_type'] == 'INV':
            y_out = y
        else:
            soften = self.task_config['label_type'] == 'soft'
            if self.task_config['task_name'] == 'grammaticality':
                # hard code for now... :(
                # 0 = ungrammatical, 1 = grammatical
                if isinstance(y, int):
                    if y == 0:
                        y_out = y
                    else: 
                        y_out = invert_label(y, soften=soften)
                else:
                    if np.argmax(y) == 0:
                        y_out = y
                    else: 
                        y_out = invert_label(y, soften=soften)
            else:
               y_out = invert_label(y, soften=soften)
        
        if self.return_metadata: 
            return X_out, y_out, metadata
        return X_out, y_out

class ChangeAntonym(ChangeSynse):
    def __init__(self, num_to_replace=-1, return_metadata=False):
        super().__init__(synse='antonym', num_to_replace=num_to_replace, return_metadata=False)
        self.return_metadata = return_metadata
        self.task_configs = [
            SentimentAnalysis(tran_type='SIB'),
            TopicClassification(),
            Grammaticality(tran_type='SIB'),
            Similarity(input_idx=[1,0], tran_type='SIB'),
            Similarity(input_idx=[0,1], tran_type='SIB'),
            Similarity(input_idx=[1,1], tran_type='INV'),
            Entailment(input_idx=[1,0], tran_type='INV'),
            Entailment(input_idx=[0,1], tran_type='INV'),
            Entailment(input_idx=[1,1], tran_type='INV'),
        ]

    def __call__(self, in_text):
        return super().__call__(in_text)

    def get_task_configs(self, task_name=None, tran_type=None, label_type=None):
        init_configs = [task() for task in self.task_configs]
        df = self._get_task_configs(init_configs, task_name, tran_type, label_type)
        return df

    def transform_Xy(self, X, y):

        # transform X
        if isinstance(X, str):
            X = [X]

        assert len(X) == len(self.task_config['input_idx']), ("The number of inputs does not match the expected "
                                                         "amount of {} for the {} task".format(
                                                            self.task_config['input_idx'],
                                                            self.task_config['task_name']))

        X_out = []
        for i, x in zip(self.task_config['input_idx'], X):
            if i == 0:
                X_out.append(x)
                continue
            X_out.append(self(x))

        metadata = {'change': X != X_out}
        X_out = X_out[0] if len(X_out) == 1 else X_out

        # transform y
        if self.task_config['tran_type'] == 'INV':
            y_out = y
        else:
            soften = self.task_config['label_type'] == 'soft'
            if self.task_config['task_name'] == 'similarity':
                # hard code for now... :(
                # 0 = dissimilar, 1 = similar
                if isinstance(y, int):
                    if y == 0:
                        y_out = 0
                    else:
                        y_out = invert_label(y, soften=soften)
                else:
                    if np.argmax(y) == 0:
                        y_out = 0
                    else:
                        y_out = smooth_label(y, factor=0.25)
            elif self.task_config['task_name'] == 'similarity':
                y_out = smooth_label(y, factor=0.5)
            else:
               y_out = invert_label(y, soften=soften)
        
        if self.return_metadata: 
            return X_out, y_out, metadata
        return X_out, y_out

class ChangeHyponym(ChangeSynse):
    def __init__(self, num_to_replace=-1, return_metadata=False):
        super().__init__(synse='hyponym', num_to_replace=num_to_replace, return_metadata=False)
        self.return_metadata = return_metadata
        self.task_configs = [
            SentimentAnalysis(),
            TopicClassification(),
            Grammaticality(tran_type='SIB'),
            Similarity(input_idx=[1,0], tran_type='INV'),
            Similarity(input_idx=[0,1], tran_type='INV'),
            Similarity(input_idx=[1,1], tran_type='INV'),
            Entailment(input_idx=[1,0], tran_type='INV'),
            Entailment(input_idx=[0,1], tran_type='INV'),
            Entailment(input_idx=[1,1], tran_type='INV'),
        ]

    def __call__(self, in_text):
        return super().__call__(in_text)

    def get_task_configs(self, task_name=None, tran_type=None, label_type=None):
        init_configs = [task() for task in self.task_configs]
        df = self._get_task_configs(init_configs, task_name, tran_type, label_type)
        return df

    def transform_Xy(self, X, y):

        # transform X
        if isinstance(X, str):
            X = [X]

        assert len(X) == len(self.task_config['input_idx']), ("The number of inputs does not match the expected "
                                                         "amount of {} for the {} task".format(
                                                            self.task_config['input_idx'],
                                                            self.task_config['task_name']))

        X_out = []
        for i, x in zip(self.task_config['input_idx'], X):
            if i == 0:
                X_out.append(x)
                continue
            X_out.append(self(x))

        metadata = {'change': X != X_out}
        X_out = X_out[0] if len(X_out) == 1 else X_out

        # transform y
        if self.task_config['tran_type'] == 'INV':
            y_out = y
        else:
            soften = self.task_config['label_type'] == 'soft'
            if self.task_config['task_name'] == 'grammaticality':
                # hard code for now... :(
                # 0 = ungrammatical, 1 = grammatical
                if isinstance(y, int):
                    if y == 0:
                        y_out = y
                    else: 
                        y_out = invert_label(y, soften=soften)
                else:
                    if np.argmax(y) == 0:
                        y_out = y
                    else: 
                        y_out = invert_label(y, soften=soften)
            elif self.task_config['task_name'] == 'similarity':
                y_out = smooth_label(y, factor=0.25)
            else:
               y_out = invert_label(y, soften=soften)
        
        if self.return_metadata: 
            return X_out, y_out, metadata
        return X_out, y_out

class ChangeHypernym(ChangeSynse):
    def __init__(self, num_to_replace=-1, return_metadata=False):
        super().__init__(synse='hypernym', num_to_replace=num_to_replace, return_metadata=False)
        self.return_metadata = return_metadata
        self.task_configs = [
            SentimentAnalysis(),
            TopicClassification(),
            Grammaticality(tran_type='SIB'),
            Similarity(input_idx=[1,0], tran_type='INV'),
            Similarity(input_idx=[0,1], tran_type='INV'),
            Similarity(input_idx=[1,1], tran_type='INV'),
            Entailment(input_idx=[1,0], tran_type='INV'),
            Entailment(input_idx=[0,1], tran_type='INV'),
            Entailment(input_idx=[1,1], tran_type='INV'),
        ]

    def __call__(self, in_text):
        return super().__call__(in_text)

    def get_task_configs(self, task_name=None, tran_type=None, label_type=None):
        init_configs = [task() for task in self.task_configs]
        df = self._get_task_configs(init_configs, task_name, tran_type, label_type)
        return df

    def transform_Xy(self, X, y):

        # transform X
        if isinstance(X, str):
            X = [X]

        assert len(X) == len(self.task_config['input_idx']), ("The number of inputs does not match the expected "
                                                         "amount of {} for the {} task".format(
                                                            self.task_config['input_idx'],
                                                            self.task_config['task_name']))

        X_out = []
        for i, x in zip(self.task_config['input_idx'], X):
            if i == 0:
                X_out.append(x)
                continue
            X_out.append(self(x))

        metadata = {'change': X != X_out}
        X_out = X_out[0] if len(X_out) == 1 else X_out

        # transform y
        if self.task_config['tran_type'] == 'INV':
            y_out = y
        else:
            soften = self.task_config['label_type'] == 'soft'
            if self.task_config['task_name'] == 'grammaticality':
                # hard code for now... :(
                # 0 = ungrammatical, 1 = grammatical
                if isinstance(y, int):
                    if y == 0:
                        y_out = y
                    else: 
                        y_out = invert_label(y, soften=soften)
                else:
                    if np.argmax(y) == 0:
                        y_out = y
                    else: 
                        y_out = invert_label(y, soften=soften)
            elif self.task_config['task_name'] == 'similarity':
                y_out = smooth_label(y, factor=0.25)
            else:
               y_out = invert_label(y, soften=soften)
        
        if self.return_metadata: 
            return X_out, y_out, metadata
        return X_out, y_out

def all_synsets(word, pos=None):
    pos_map = {
        'NOUN': wordnet.NOUN,
        'VERB': wordnet.VERB,
        'ADJ': wordnet.ADJECTIVE,
        'ADV': wordnet.ADVERB
        }
    if pos is None:
        pos_list = [wordnet.VERB, wordnet.ADJECTIVE, wordnet.NOUN, wordnet.ADVERB]
    else:
        pos_list = [pos_map[pos]]
    out_text = []
    for pos in pos_list:
        out_text.extend(wordnet.synsets(word, pos=pos))
    return out_text

def clean_senses(synsets):
    return [x for x in set(synsets) if '_' not in x]

def all_possible_synonyms(word, pos=None):
    out_text = []
    for syn in all_synsets(word, pos=pos):
        # if syn.synonyms[0] != word:
        #     continue
        out_text.extend(syn.senses)
    return clean_senses(out_text)

def all_possible_antonyms(word, pos=None):
    out_text = []
    for syn in all_synsets(word, pos=pos):
        if not syn.antonym:
            continue
        for s in syn.antonym:
            out_text.extend(s.senses)            
    if not out_text:
        synos = all_possible_synonyms(word, pos)
        for syno in synos:
            for syn in all_synsets(syno, pos=pos):
                if not syn.antonym:
                    continue
                for s in syn.antonym:
                    out_text.extend(s.senses)
    return clean_senses(out_text)

def all_possible_hypernyms(word, pos=None, depth=3):
    out_text = []
    for syn in all_synsets(word, pos=pos):
        out_text.extend([y for x in syn.hypernyms(recursive=True, depth=depth) for y in x.senses])
    return clean_senses(out_text)

def all_possible_hyponyms(word, pos=None, depth=3):
    out_text = []
    for syn in all_synsets(word, pos=pos):
        out_text.extend([y for x in syn.hyponyms(recursive=True, depth=depth) for y in x.senses])
    return clean_senses(out_text)

def strip_punct(word):
    puncts = re.compile(r'[^\w\s]')
    return puncts.sub('', word)


__all__ = [
	'ExpandContractions',
	'ContractContractions',
	'Emojify',
	'AddEmoji',
	'AddPositiveEmoji',
	'AddNegativeEmoji',
	'AddNeutralEmoji',
	'Demojify', 
	'RemoveEmoji', 
	'RemovePositiveEmoji',
	'RemoveNegativeEmoji',
	'RemoveNeutralEmoji',
	'ChangeLocation',
	'ChangeName',
	'InsertPositivePhrase',
	'InsertNegativePhrase',
	'RandomInsertion',
	'AddSentimentLink',
	'ImportLinkText',
	'AddNegation',
	'RemoveNegation',
	'RandomCharDel',
	'RandomCharInsert', 
	'RandomCharSubst', 
	'RandomCharSwap', 
	'RandomSwapQwerty',
	'ChangeNumber',
	'ChangeSynonym', 
	'ChangeAntonym', 
	'ChangeHyponym', 
	'ChangeHypernym',
	'WordDeletion', 
	'HomoglyphSwap', 
	'RandomSwap'
]


from ..abstract_transformation import *
from ..tasks import *
from ..data.locations import NAMED_ENTITIES
import numpy as np
import en_core_web_sm

class ChangeLocation(AbstractTransformation):
    """
    Transforms an input by replacing names of recognized 
    location entity.
    """

    def __init__(self, return_metadata=False):
        """
        Parameters
        ----------
        return_metadata : bool
            whether or not to return metadata, e.g. 
            whether a transform was successfully
            applied or not
        """
        super().__init__() 
        self.nlp = en_core_web_sm.load()
        self.return_metadata = return_metadata
        self.task_configs = [
            SentimentAnalysis(),
            TopicClassification(),
            Grammaticality(),
            Similarity(input_idx=[1,0], tran_type='SIB'),
            Similarity(input_idx=[0,1], tran_type='SIB'),
            Similarity(input_idx=[1,1], tran_type='SIB'),
            Entailment(input_idx=[1,0], tran_type='SIB'),
            Entailment(input_idx=[0,1], tran_type='SIB'),
            Entailment(input_idx=[1,1], tran_type='SIB'),
        ]
    
    def __call__(self, in_text):
        doc = self.nlp(in_text)
        out_text = in_text
        for e in reversed(doc.ents): #reversed to not modify the offsets of other entities when substituting
            start = e.start_char
            end = start + len(e.text)
            # print(e.text, "label is ", e.label_)
            if e.label_ in ('GPE', 'NORP', 'FAC', 'ORG', 'LOC'):
                name = out_text[start:end]
                name = name.split()
                name[0] =  self._get_loc_name()
                name = " ".join(name)
                out_text = out_text[:start] + name + out_text[end:]
        return out_text

    def _get_loc_name(self):
        """Return a random location name."""
        loc = self.np_random.choice(['country', 'nationality', 'city'])
        return self.np_random.choice(NAMED_ENTITIES[loc])

    def get_task_configs(self, task_name=None, tran_type=None, label_type=None):
        init_configs = [task() for task in self.task_configs]
        df = self._get_task_configs(init_configs, task_name, tran_type, label_type)
        return df
        
    def transform_Xy(self, X, y):

        # transform X
        if isinstance(X, str):
            X = [X]

        assert len(X) == len(self.task_config['input_idx']), ("The number of inputs does not match the expected "
                                                         "amount of {} for the {} task".format(
                                                            self.task_config['input_idx'],
                                                            self.task_config['task_name']))

        X_out = []
        for i, x in zip(self.task_config['input_idx'], X):
            if i == 0:
                X_out.append(x)
                continue
            X_out.append(self(x))

        metadata = {'change': X != X_out}
        X_out = X_out[0] if len(X_out) == 1 else X_out

        # transform y
        if self.task_config['tran_type'] == 'INV':
            y_out = y
        else:
            soften = self.task_config['label_type'] == 'soft'
            if self.task_config['task_name'] == 'similarity':
                # hard code for now... :(
                # 0 = dissimilar, 1 = similar
                if isinstance(y, int):
                    if y == 0:
                        y_out = 0
                    else:
                        y_out = invert_label(y, soften=soften)
                else:
                    if np.argmax(y) == 0:
                        y_out = 0
                    else:
                        y_out = smooth_label(y, factor=0.25)
            elif self.task_config['task_name'] == 'entailment':
                # hard coded for now... :(
                # 0 = entailed, 1 = neutral, 2 = contradiction
                if isinstance(y, int):
                    if y in [0, 2]:
                        y_out = 1
                    else: 
                        y_out = y
                else:
                    if np.argmax(y) in [0, 2]:
                        y_out = 1
                    else:
                        y_out = y
            else:
                y_out = invert_label(y, soften=soften)
        
        if self.return_metadata: 
            return X_out, y_out, metadata
        return X_out, y_out



from ..abstract_batch_transformation import AbstractBatchTransformation
from ...utils import one_hot_encode
from ..tasks import *
import numpy as np
import pandas as pd
import itertools
import nltk

try:
    nltk.data.find('tokenizers/punkt')
except LookupError:
    nltk.download('punkt')

class TextMix(AbstractBatchTransformation):
    """
    Concatenates two texts together and interpolates 
    the labels
    """
    def __init__(self, return_metadata=False):
        """
        Initializes the transformation and provides an
        opporunity to supply a configuration if needed

        Parameters
        ----------
        return_metadata : bool
            whether or not to return metadata, e.g. 
            whether a transform was successfully
            applied or not
        """
        super().__init__() 
        self.return_metadata = return_metadata
        self.task_configs = [
            SentimentAnalysis(tran_type='SIB'),
            TopicClassification(tran_type='SIB'),
            Grammaticality(tran_type='SIB'),
            # Similarity(input_idx=[1,0], tran_type='SIB'),
            # Similarity(input_idx=[0,1], tran_type='SIB'),
            # Similarity(input_idx=[1,1], tran_type='SIB'),
            # Entailment(input_idx=[1,0], tran_type='SIB'),
            # Entailment(input_idx=[0,1], tran_type='SIB'),
            # Entailment(input_idx=[1,1], tran_type='SIB'),
        ]
        
    def __call__(self, batch, target_pairs=[], target_prob=0, num_classes=2):
        """
        Parameters
        ----------
        batch : tuple(np.array, np.array)
            The input data and the targets, which will
            be one-hot-encoded if they aren't already
        target_pairs : list[tuple(int, int)]
            A list of tuple pairs that represent a mapping
            between a source_class and target_class. The 
            target_class gets cut into the source_class. 
            For all unspecified class pairings, a random 
            pairing will be made
        target_prob : float [0,1]
            The probability of applying the targeted cutmix
            for a particular pairing
        num_classes : int
            The number of classes in the dataset

        Returns
        ----------
        ret : tuple(np.array, np.array)
            : tuple(np.array, np.array, dict)
            The output batch of transformed (X, y)
            pairs. y --> one-hot-encoded, weighted
            by the amount of text drawn from each
            of the inputs. May return metadata dict
            if requested.
        """
        
        # unpack batch
        data, targets = batch
        batch_size = len(data)

        # convert to numpy if not already
        if type(data) == list:
            data = [str(x).encode('utf-8') for x in data]
            data = np.array(data, dtype=np.string_)
        if type(targets) == list:
            if type(targets[0]) == np.ndarray:
                targets = np.stack(targets)
            else:
                targets = np.array(targets)

        # track indices for targeted cutmix to exclude later
        idx = [x for x in range(batch_size)]
        ex_idx = []

        new_data = []
        new_targets = []

        # transform targeted pairings
        for pair in target_pairs:
            
            # skip targeted transformation target_prob percent of the time
            use_targets = self.np_random.uniform() < target_prob
            if not use_targets:
                continue
                
            # unpack source and target pairs
            source_class, target_class = pair

            # find indices of both source and target 
            s_idx = find_value_idx(targets, source_class)
            t_idx = find_value_idx(targets, target_class)
            
            # if none of the source or target classes are in this batch, skip it
            if len(s_idx) == 0 or len(t_idx) == 0:
                continue
                
            # enforce source==target array size via sampling
            tt_idx = self.np_random.choice(np.arange(len(t_idx)), size=len(s_idx), replace=True)
            t_idx = t_idx[tt_idx]
            
            # create concatenated data
            textmix = concat_text(data[s_idx], data[t_idx])
            ohe_targets = concat_labels(data[s_idx], 
                                        data[t_idx], 
                                        targets[s_idx],
                                        targets[t_idx],
                                        num_classes)
            
            new_data.append(textmix)
            new_targets.append(ohe_targets)
            ex_idx.append(s_idx.tolist())
            
        ex_idx = list(itertools.chain(*ex_idx))
        s_idx = [i for i in idx if i not in ex_idx]
        t_idx = self.np_random.choice(np.arange(len(s_idx)), size=len(s_idx), replace=True)

        if s_idx:
            textmix = concat_text(data[s_idx], data[t_idx])
            ohe_targets = concat_labels(data[s_idx], 
                                        data[t_idx], 
                                        targets[s_idx],
                                        targets[t_idx],
                                        num_classes)

            new_data.append(textmix)
            new_targets.append(ohe_targets)
            
        new_data = np.concatenate(new_data)
        new_data = [x.decode() if type(x) == bytes else str(x) for x in new_data]
        new_targets = np.concatenate(new_targets).tolist()

        out_text = (new_data, new_targets)

        # metadata
        if self.return_metadata: 
            metadata = {'change': True}
            return out_text, metadata
        return out_text

    def get_task_configs(self, task_name=None, tran_type=None, label_type=None):
        init_configs = [task() for task in self.task_configs]
        df = self._get_task_configs(init_configs, task_name, tran_type, label_type)
        return df


class SentMix(AbstractBatchTransformation):
    """
    Concatenates two texts together and then mixes the
    sentences in the new string. Interpolates the labels
    """
    def __init__(self, return_metadata=False):
        """
        Initializes the transformation and provides an
        opporunity to supply a configuration if needed

        Parameters
        ----------
        return_metadata : bool
            whether or not to return metadata, e.g. 
            whether a transform was successfully
            applied or not
        """
        super().__init__() 
        self.return_metadata = return_metadata
        self.task_configs = [
            SentimentAnalysis(tran_type='SIB'),
            TopicClassification(tran_type='SIB'),
            Grammaticality(tran_type='SIB'),
            # Similarity(input_idx=[1,0], tran_type='SIB'),
            # Similarity(input_idx=[0,1], tran_type='SIB'),
            # Similarity(input_idx=[1,1], tran_type='SIB'),
            # Entailment(input_idx=[1,0], tran_type='SIB'),
            # Entailment(input_idx=[0,1], tran_type='SIB'),
            # Entailment(input_idx=[1,1], tran_type='SIB'),
        ]
        
    def __call__(self, batch, target_pairs=[], target_prob=0, num_classes=2):
        """
        Parameters
        ----------
        batch : tuple(np.array, np.array)
            The input data and the targets, which will
            be one-hot-encoded if they aren't already
        target_pairs : list[tuple(int, int)]
            A list of tuple pairs that represent a mapping
            between a source_class and target_class. The 
            target_class gets cut into the source_class. 
            For all unspecified class pairings, a random 
            pairing will be made
        target_prob : float [0,1]
            The probability of applying the targeted cutmix
            for a particular pairing
        num_classes : int
            The number of classes in the dataset

        Returns
        ----------
        ret : tuple(np.array, np.array)
            : tuple(np.array, np.array, dict)
            The output batch of transformed (X, y)
            pairs. y --> one-hot-encoded, weighted
            by the amount of text drawn from each
            of the inputs. May return metadata dict
            if requested.
        """

        # TextMix transformation
        new_data, new_targets = TextMix()(batch, target_pairs, target_prob, num_classes)

        # mix sentences
        sent_shuffle_ = np.vectorize(self.sent_shuffle)
        new_data = np.apply_along_axis(sent_shuffle_, 0, new_data).tolist()

        out_text = (new_data, new_targets)

        # metadata
        if self.return_metadata: 
            metadata = {'change': True}
            return out_text, metadata
        return out_text

    def sent_shuffle(self, string):
        X = nltk.tokenize.sent_tokenize(str(string))
        self.np_random.shuffle(X)
        return ' '.join(X)

    def get_task_configs(self, task_name=None, tran_type=None, label_type=None):
        init_configs = [task() for task in self.task_configs]
        df = self._get_task_configs(init_configs, task_name, tran_type, label_type)
        return df

class WordMix(AbstractBatchTransformation):
    """
    Concatenates two texts together and then mixes the
    words in the new string. Interpolates the labels
    """
    def __init__(self, return_metadata=False):
        """
        Initializes the transformation and provides an
        opporunity to supply a configuration if needed

        Parameters
        ----------
        return_metadata : bool
            whether or not to return metadata, e.g. 
            whether a transform was successfully
            applied or not
        """
        super().__init__() 
        self.return_metadata = return_metadata
        self.task_configs = [
            SentimentAnalysis(tran_type='SIB'),
            TopicClassification(tran_type='SIB'),
            Grammaticality(tran_type='SIB'),
            # Similarity(input_idx=[1,0], tran_type='SIB'),
            # Similarity(input_idx=[0,1], tran_type='SIB'),
            # Similarity(input_idx=[1,1], tran_type='SIB'),
            # Entailment(input_idx=[1,0], tran_type='SIB'),
            # Entailment(input_idx=[0,1], tran_type='SIB'),
            # Entailment(input_idx=[1,1], tran_type='SIB'),
        ]
        
    def __call__(self, batch, target_pairs=[], target_prob=0, num_classes=2):
        """
        Parameters
        ----------
        batch : tuple(np.array, np.array)
            The input data and the targets, which will
            be one-hot-encoded if they aren't already
        target_pairs : list[tuple(int, int)]
            A list of tuple pairs that represent a mapping
            between a source_class and target_class. The 
            target_class gets cut into the source_class. 
            For all unspecified class pairings, a random 
            pairing will be made
        target_prob : float [0,1]
            The probability of applying the targeted cutmix
            for a particular pairing
        num_classes : int
            The number of classes in the dataset

        Returns
        ----------
        ret : tuple(np.array, np.array)
            : tuple(np.array, np.array, dict)
            The output batch of transformed (X, y)
            pairs. y --> one-hot-encoded, weighted
            by the amount of text drawn from each
            of the inputs. May return metadata dict
            if requested.
        """

        # TextMix transformation
        new_data, new_targets = TextMix()(batch, target_pairs, target_prob, num_classes)

        # mix words
        word_shuffle_ = np.vectorize(self.word_shuffle)
        new_data = np.apply_along_axis(word_shuffle_, 0, new_data).tolist()

        out_text = (new_data, new_targets)

        # metadata
        if self.return_metadata: 
            metadata = {'change': True}
            return out_text, metadata
        return out_text

    def word_shuffle(self, string):
        X = str(string).split(" ")
        self.np_random.shuffle(X)
        return ' '.join(X)

    def get_task_configs(self, task_name=None, tran_type=None, label_type=None):
        init_configs = [task() for task in self.task_configs]
        df = self._get_task_configs(init_configs, task_name, tran_type, label_type)
        return df

# Helper functions
def concat_labels(source_text, 
                  target_text, 
                  source_labels, 
                  target_labels, 
                  num_classes):
    
    # create soft target labels 
    source_ohe = one_hot_encode(source_labels, num_classes)
    target_ohe = one_hot_encode(target_labels, num_classes)

    if len(source_ohe.shape) == 1:
        source_ohe = np.expand_dims(source_ohe, 0)
    if len(target_ohe.shape) == 1:
        target_ohe = np.expand_dims(target_ohe, 0)
    
    if source_labels.shape[-1] == 1:
        source_cls = source_labels
        target_cls = target_labels
    else:
        source_cls = np.argmax(source_ohe, axis=1)
        target_cls = np.argmax(target_ohe, axis=1)
        
    # calculate length of each data and use that
    # to determine the lambda weight assigned to
    # the index for the target
    len_data_source = np.char.str_len(source_text)
    len_data_target = np.char.str_len(target_text)
    lam = len_data_source / (len_data_source + len_data_target)   

    idx_ = np.arange(len(source_ohe))

    # print(lam.shape, lam)
    # print(idx_.shape, idx_)
    # print(source_ohe.shape, source_ohe)
    # print(target_ohe.shape, target_ohe)
    # print(source_cls.shape, source_cls)
    # print(target_cls.shape, target_cls)
    
    # NOTE: https://stackoverflow.com/questions/38673531/numpy-cannot-cast-ufunc-multiply-output-from-dtype
    source_ohe[idx_, source_cls] = (source_ohe[idx_, source_cls] * lam)
    target_ohe[idx_, target_cls] = (target_ohe[idx_, target_cls] * 1-lam)
    
    ohe_targets = source_ohe + target_ohe
    
    return ohe_targets

def concat_text(np_char1, np_char2):
    np_char1 = np_char1.astype(np.string_)
    np_char2 = np_char2.astype(np.string_)
    sep = np.full_like(np_char1, " ", dtype=np.string_)
    out_text = np.char.add(np_char1, sep)
    out_text = np.char.add(out_text, np_char2)
    return out_text

def find_value_idx(array, value):
    return np.asarray(array == value).nonzero()[0]

def find_other_idx(array, value):
    return np.asarray(array != value).nonzero()[0]


from ..abstract_batch_transformation import AbstractBatchTransformation
from ...utils import one_hot_encode
from ..tasks import *
from ..generative.concept2sentence import Concept2Sentence
import numpy as np
import pandas as pd
import itertools
import nltk

class ConceptMix(AbstractBatchTransformation):

    uses_dataset = True
    
    def __init__(self, 
                 return_metadata=False,
                 dataset=None, 
                 extract="token",
                 gen_beam_size=10,
                 text_min_length=10,
                 text_max_length=32,
                 device='cuda',
                 generation_type='disjoint',
                 task_config=None):
        """
        Initializes the transformation and provides an
        opporunity to supply a configuration if needed

        Parameters
        ----------
        return_metadata : bool
            whether or not to return metadata, e.g. 
            whether a transform was successfully
            applied or not
        """
        super().__init__() 
        self.return_metadata = return_metadata
        self.task_configs = [
            SentimentAnalysis(tran_type='SIB'),
            TopicClassification(tran_type='SIB'),
            Grammaticality(),
            # Similarity(input_idx=[1,0], tran_type='SIB'),
            # Similarity(input_idx=[0,1], tran_type='SIB'),
            # Similarity(input_idx=[1,1], tran_type='SIB'),
            # Entailment(input_idx=[1,0], tran_type='SIB'),
            # Entailment(input_idx=[0,1], tran_type='SIB'),
            # Entailment(input_idx=[1,1], tran_type='SIB'),
        ]
        self.c2s = Concept2Sentence(
                        dataset = dataset,
                        extract = extract,
                        gen_beam_size = gen_beam_size,
                        text_min_length = text_min_length,
                        text_max_length = text_max_length,
                        device = device)
        self.generation_type = generation_type
        if self.generation_type not in ['joint', 'disjoint']:
            raise ValueError("Must select a generation type from ['joint', 'disjoin']")
        
    def __call__(self, batch, target_pairs=[], target_prob=1, num_classes=2):
        """
        Parameters
        ----------
        batch : tuple(np.array, np.array)
            The input data and the targets, which will
            be one-hot-encoded if they aren't already
        target_pairs : list[tuple(int, int)]
            A list of tuple pairs that represent a mapping
            between a source_class and target_class. The 
            target_class gets cut into the source_class. 
            For all unspecified class pairings, a random 
            pairing will be made
        target_prob : float [0,1]
            The probability of applying the targeted mixing
            for a particular pairing
        num_classes : int
            The number of classes in the dataset

        Returns
        ----------
        ret : tuple(np.array, np.array)
            : tuple(np.array, np.array, dict)
            The output batch of transformed (X, y)
            pairs. y --> one-hot-encoded, weighted
            by the amount of text drawn from each
            of the inputs. May return metadata dict
            if requested.
        """
        
        # unpack batch
        data, targets = batch
        batch_size = len(data)

        # convert to numpy if not already
        if type(data) == list:
            data = [str(x).encode('utf-8') for x in data]
            data = np.array(data, dtype=np.string_)
        if type(targets) == list:
            if type(targets[0]) == np.ndarray:
                targets = np.stack(targets)
            else:
                targets = np.array(targets)

        # track indices for targeted cutmix to exclude later
        idx = [x for x in range(batch_size)]
        ex_idx = []

        new_data = []
        new_targets = []

        # transform targeted pairings
        for pair in target_pairs:
            
            # skip targeted transformation target_prob percent of the time
            use_targets = self.np_random.uniform() < target_prob
            if not use_targets:
                continue
                
            # unpack source and target pairs
            source_class, target_class = pair

            # find indices of both source and target 
            s_idx = find_value_idx(targets, source_class)
            t_idx = find_value_idx(targets, target_class)
            
            # if none of the source or target classes are in this batch, skip it
            if len(s_idx) == 0 or len(t_idx) == 0:
                continue
                
            # enforce source==target array size via sampling
            tt_idx = self.np_random.choice(np.arange(len(t_idx)), size=len(s_idx), replace=True)
            t_idx = t_idx[tt_idx]
            
            # create concatenated data
            new_text, new_target = self.conmix(data[s_idx], 
                                               data[t_idx], 
                                               targets[s_idx],
                                               targets[t_idx],
                                               num_classes)
            
            new_data.append(new_text)
            new_targets.append(new_target)
            ex_idx.append(s_idx.tolist())
            
        ex_idx = list(itertools.chain(*ex_idx))
        s_idx = [i for i in idx if i not in ex_idx]
        t_idx = self.np_random.choice(np.arange(len(s_idx)), size=len(s_idx), replace=True)

        if s_idx:
            new_text, new_target = self.conmix(data[s_idx], 
                                               data[t_idx], 
                                               targets[s_idx],
                                               targets[t_idx],
                                               num_classes)

            new_data.append(new_text)
            new_targets.append(new_target)
            
        new_data = np.concatenate(new_data)
        new_data = [x.decode() if type(x) == bytes else str(x) for x in new_data]
        new_targets = np.concatenate(new_targets).tolist()

        out = (new_data, new_targets)

        # metadata
        if self.return_metadata: 
            metadata = {'change': True}
            return out, metadata
        return out

    def get_task_configs(self, task_name=None, tran_type=None, label_type=None):
        init_configs = [task() for task in self.task_configs]
        df = self._get_task_configs(init_configs, task_name, tran_type, label_type)
        return df

    def conmix(self, texts1, texts2, targets1, targets2, num_classes):

        # prep text
        texts1 = texts1.astype(np.string_)
        texts2 = texts2.astype(np.string_)
        # prep targets
        targets1_ohe = np.array([one_hot_encode(t, num_classes).astype(np.float32) for t in targets1])
        targets2_ohe = np.array([one_hot_encode(t, num_classes).astype(np.float32) for t in targets2])
        if len(targets1_ohe.shape) == 1:
            targets1_ohe = np.expand_dims(targets1_ohe, 0)
        if len(targets2_ohe.shape) == 1:
            targets2_ohe = np.expand_dims(targets2_ohe, 0)
        if targets1.shape[-1] == 1:
            targets1_cls = targets1
            targets2_cls = targets2
        else:
            targets1_cls = np.argmax(targets1_ohe, axis=1)
            targets2_cls = np.argmax(targets2_ohe, axis=1)

        # print('texts1', texts1)
        # print('texts2', texts2)
        # print('targets1', targets1)
        # print('targets2', targets2)
        # print('targets1_ohe', targets1_ohe)
        # print('targets2_ohe', targets2_ohe)
        # print('targets1_cls', targets1_cls)
        # print('targets2_cls', targets2_cls)

        # transform
        new_texts, new_targets = [], []
        for s1, s2, t1_ohe, t2_ohe, t1_cls, t2_cls in zip(texts1, texts2, 
                                                          targets1_ohe, targets2_ohe, 
                                                          targets1_cls, targets2_cls):

            if isinstance(s1, bytes):
                s1 = s1.decode()
            if isinstance(s2, bytes):
                s2 = s2.decode()

            # transform text
            concepts1 = self.c2s.extract_concepts(s1, t1_cls)
            concepts2 = self.c2s.extract_concepts(s2, t2_cls)
            combined_concepts = list(set(concepts1 + concepts2))

            if not combined_concepts:
                # if there aren't any concepts extracted
                # simply append the original sentence and target unchanged
                new_texts.append(s1)
                new_targets.append(t1_ohe)

            else:
                if self.generation_type == 'joint':

                    new_text = self.c2s.generate_text_from_concepts(combined_concepts)

                elif self.generation_type == 'disjoint':

                    if concepts1:
                        new_text1 = self.c2s.generate_text_from_concepts(concepts1)
                    else:
                        new_text1 = s1

                    if not concepts2:
                        new_text2 = self.c2s.generate_text_from_concepts(concepts2)
                    else:
                        new_text2 = s2

                    new_text = new_text1 + " " + new_text2

                # transform targets
                lam = len(concepts1) / len(combined_concepts)
                t1_ohe[t1_cls] *= lam
                t2_ohe[t2_cls] *= 1-lam    
                new_target = np.array(t1_ohe) + np.array(t2_ohe)

                new_texts.append(new_text)
                new_targets.append(new_target)

        return new_texts, new_targets

# Helper functions

def find_value_idx(array, value):
    return np.asarray(array == value).nonzero()[0]

def find_other_idx(array, value):
    return np.asarray(array != value).nonzero()[0]

# if __name__ == '__main__':
#     import json
#     tf = ConceptMix(dataset='sst2')
#     texts = ["I hate how long loading the models takes to select better keyphrases.",
#              "I really love this movie a lot!"]
#     targets = [0, 1]
#     batch = (texts, targets)
#     results = []
#     new_text, new_target = tf(batch, num_classes=2)
#     results.append({
#         "class": tf.name(),
#         "inputs": {"texts": texts, "targets": targets}, 
#         "outputs": {"new_text": new_text, "new_target": new_target}
#     })
#     json_file = {"type": tf.name(), "results": results}
#     print(json.dumps(json_file, indent=2))


from ..abstract_transformation import *
from ..tasks import *
import collections
import pattern
import spacy
import en_core_web_sm
import numpy as np

class RemoveNegation(AbstractTransformation):
    """
    Defines a transformation that removes a negation
    if found within the input. Otherwise, return the 
    original string unchanged. 
    """

    def __init__(self, return_metadata=False):
        """
        Initializes the transformation and provides an
        opporunity to supply a configuration if needed

        Parameters
        ----------
        return_metadata : bool
            whether or not to return metadata, e.g. 
            whether a transform was successfully
            applied or not
        """
        super().__init__() 
        self.nlp = en_core_web_sm.load()
        self.return_metadata = return_metadata
        self.task_configs = [
            SentimentAnalysis(),
            TopicClassification(),
            Grammaticality(),
            Similarity(input_idx=[1,0], tran_type='SIB'),
            Similarity(input_idx=[0,1], tran_type='SIB'),
            Similarity(input_idx=[1,1], tran_type='INV'),
            Entailment(input_idx=[1,0], tran_type='SIB'),
            Entailment(input_idx=[0,1], tran_type='SIB'),
            Entailment(input_idx=[1,1], tran_type='INV'),
        ]
    
    def __call__(self, in_text):
        # This removes all negations in the doc. I should maybe add an option to remove just some.
        doc = self.nlp(in_text)
        doc_len = len(doc)
        notzs = [i for i, z in enumerate(doc) if z.lemma_ == 'not' or z.dep_ == 'neg']
        new = []
        for notz in notzs:
            before = doc[notz - 1] if notz != 0 else None
            after = doc[notz + 1] if len(doc) > notz + 1 else None
            if (after and after.pos_ == 'PUNCT') or (before and before.text in ['or']):
                continue
            new.append(notz)
        notzs = new
        if not notzs:
            return in_text
        out_text = ''
        start = 0
        for i, notz in enumerate(notzs):
            id_start = notz
            to_add = ' '
            id_end = notz + 1
            before = doc[notz - 1] if notz != 0 else None
            after = doc[notz + 1] if len(doc) > notz + 1 else None
            if before and before.lemma_ in ['will', 'can', 'do']:
                id_start = notz - 1
                tenses = collections.Counter([x[0] for x in pattern.en.tenses(before.text)]).most_common(1)
                tense = tenses[0][0] if len(tenses) else 'present'
                p = pattern.en.tenses(before.text)
                params = [tense, 3]
                if p:
                    tmp = [x for x in p if x[0] == tense]
                    if tmp:
                        params = list(tmp[0])
                    else:
                        params = list(p[0])
                to_add = ' '+ pattern.en.conjugate(before.lemma_, *params) + ' '
            if before and after and before.lemma_ == 'do' and after.pos_ == 'VERB':
                id_start = notz - 1
                tenses = collections.Counter([x[0] for x in pattern.en.tenses(before.text)]).most_common(1)
                tense = tenses[0][0] if len(tenses) else 'present'
                p = pattern.en.tenses(before.text)
                params = [tense, 3]
                if p:
                    tmp = [x for x in p if x[0] == tense]
                    if tmp:
                        params = list(tmp[0])
                    else:
                        params = list(p[0])
                to_add = ' '+ pattern.en.conjugate(after.text, *params) + ' '
                id_end = notz + 2
            # print(doc)
            # print(start)
            # print(id_start)
            # print(to_add)
            if start == id_start or start >= doc_len or id_start >= doc_len:
                continue
            else:
                out_text += doc[start:id_start].text + to_add
            start = id_end
        if id_end < doc_len:
            out_text += doc[id_end:].text
        return out_text

    def get_task_configs(self, task_name=None, tran_type=None, label_type=None):
        init_configs = [task() for task in self.task_configs]
        df = self._get_task_configs(init_configs, task_name, tran_type, label_type)
        return df
        
    def transform_Xy(self, X, y):

        # transform X
        if isinstance(X, str):
            X = [X]

        assert len(X) == len(self.task_config['input_idx']), ("The number of inputs does not match the expected "
                                                         "amount of {} for the {} task".format(
                                                            self.task_config['input_idx'],
                                                            self.task_config['task_name']))

        X_out = []
        for i, x in zip(self.task_config['input_idx'], X):
            if i == 0:
                X_out.append(x)
                continue
            X_out.append(self(x))

        metadata = {'change': X != X_out}
        X_out = X_out[0] if len(X_out) == 1 else X_out

        # transform y
        if self.task_config['tran_type'] == 'INV':
            y_out = y
        else:                
            soften = self.task_config['label_type'] == 'soft'
            if self.task_config['task_name'] == 'similarity':
                # hard code for now... :(
                # 0 = dissimilar, 1 = similar
                if isinstance(y, int):
                    if y == 0:
                        y_out = 0
                    else:
                        y_out = invert_label(y, soften=soften)
                else:
                    if np.argmax(y) == 0:
                        y_out = 0
                    else:
                        y_out = smooth_label(y, factor=0.25)
            elif self.task_config['task_name'] == 'entailment':
                # hard coded for now... :(
                # 0 = entailed, 1 = neutral, 2 = contradiction
                if isinstance(y, int):
                    if y in [0, 2]:
                        y_out = 1
                    else: 
                        y_out = y
                else:
                    if np.argmax(y) in [0, 2]:
                        y_out = 1
                    else:
                        y_out = y
            else:
                y_out = invert_label(y, soften=soften)
        
        if self.return_metadata: 
            return X_out, y_out, metadata
        return X_out, y_out


from ..abstract_transformation import *
from ..tasks import *

class WordDeletion(AbstractTransformation):
    """
    Deletes words from random indices in the string input
    """

    def __init__(self, p=0.25, return_metadata=False):
        """
        Initializes the transformation and provides an
        opporunity to supply a configuration if needed

        Parameters
        ----------
        p : float
            Randomly delete words from the sentence with probability p
        return_metadata : bool
            whether or not to return metadata, e.g. 
            whether a transform was successfully
            applied or not
        """
        super().__init__() 
        self.p = p
        self.return_metadata = return_metadata
        self.task_configs = [
            SentimentAnalysis(),
            TopicClassification(),
            Grammaticality(tran_type='SIB'),
            Similarity(input_idx=[1,0], tran_type='INV'),
            Similarity(input_idx=[0,1], tran_type='INV'),
            Similarity(input_idx=[1,1], tran_type='INV'),
            Entailment(input_idx=[1,0], tran_type='INV'),
            Entailment(input_idx=[0,1], tran_type='INV'),
            Entailment(input_idx=[1,1], tran_type='INV'),
        ]
    
    def __call__(self, in_text):
        """
        Parameters
        ----------
        in_text : str
            The input string

        Returns
        ----------
        out_text : str
            The output with random words deleted
        """
        words = in_text.split()
        # obviously, if there's only one word, don't delete it
        if len(words) == 1:
            return in_text

        #randomly delete words with probability p
        new_words = []
        for word in words:
            r = self.np_random.uniform(0, 1)
            if r > self.p:
                new_words.append(word)

        #if you end up deleting all words, just return a random word
        if len(new_words) == 0:
            rand_int = self.np_random.integers(0, len(words)-1)
            return words[rand_int]

        out_text = ' '.join(new_words)
        return out_text

    def get_task_configs(self, task_name=None, tran_type=None, label_type=None):
        init_configs = [task() for task in self.task_configs]
        df = self._get_task_configs(init_configs, task_name, tran_type, label_type)
        return df

    def transform_Xy(self, X, y):

        # transform X
        if isinstance(X, str):
            X = [X]

        assert len(X) == len(self.task_config['input_idx']), ("The number of inputs does not match the expected "
                                                         "amount of {} for the {} task".format(
                                                            self.task_config['input_idx'],
                                                            self.task_config['task_name']))

        X_out = []
        for i, x in zip(self.task_config['input_idx'], X):
            if i == 0:
                X_out.append(x)
                continue
            X_out.append(self(x))

        metadata = {'change': X != X_out}
        X_out = X_out[0] if len(X_out) == 1 else X_out

        # transform y
        if self.task_config['tran_type'] == 'INV':
            y_out = y
        else:
            soften = self.task_config['label_type'] == 'soft'
            if self.task_config['task_name'] == 'grammaticality':
                # hard code for now... :(
                # 0 = ungrammatical, 1 = grammatical
                if isinstance(y, int):
                    if y == 0:
                        y_out = y
                    else: 
                        y_out = invert_label(y, soften=soften)
                else:
                    if np.argmax(y) == 0:
                        y_out = y
                    else: 
                        y_out = invert_label(y, soften=soften)
            else:
                y_out = invert_label(y, soften=soften)
        
        if self.return_metadata: 
            return X_out, y_out, metadata
        return X_out, y_out